def create_signed_dockerfile(image: str, signature_label: str, signature: str) -> str:
    dockerfile_template = """
        FROM {image_name}
        LABEL {signature_label}={signature}
        """.format(
            image_name=image, signature_label=signature_label, signature=signature
        )

    return dockerfile_template



from cryptography.hazmat.primitives.serialization import (
    load_pem_public_key,
    load_pem_private_key,
)
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature

from notarizer.converter import str_to_bytes, bytes_to_str, encode_to_base64
from notarizer.exceptions import VerificationFailure


def verify(public_key: bytes, signature: bytes, content: str):
    try:
        public_key_pem = load_pem_public_key(public_key, default_backend())
        public_key_pem.verify(
            signature,
            str_to_bytes(content),
            padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
            hashes.SHA256(),
        )
    except InvalidSignature:
        raise VerificationFailure("Invalid Signature")


def create_signature(private_key: bytes, content: str) -> str:
    private_key_pem = load_pem_private_key(private_key, password=None, backend=default_backend())

    signature = private_key_pem.sign(
        str_to_bytes(content),
        padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH),
        hashes.SHA256(),
    )

    encoded_signature = encode_to_base64(signature)
    signature_text = bytes_to_str(encoded_signature)

    return signature_text



class MockedData:
    @staticmethod
    def get_public_key_bytes():
        return bytes("dummyPublicKey", "utf-8")

    @staticmethod
    def get_private_key_bytes():
        return bytes("dummyPrivateKey", "utf-8")

    @staticmethod
    def get_fake_valid_history_content(label):
        history = [
            {"CreatedBy": "/bin/sh -c #(nop)  LABEL {label}=dummyb64".format(label=label)},
            {"CreatedBy": '/bin/sh -c #(nop)  CMD ["/hello"]'},
            {"CreatedBy": "/bin/sh -c #(nop) COPY file:dummysha in / "},
        ]
        return [hline["CreatedBy"] for hline in history]

    @staticmethod
    def get_fake_invalid_history_content(label):
        history = [
            {"CreatedBy": "/bin/sh -c #(nop)  LABEL {label}=dummyb64".format(label=label)},
            {"CreatedBy": "/bin/sh -c #(nop) COPY file:dummysha in / "},
        ]
        return [hline["CreatedBy"] for hline in history]



from base64 import b64encode, decodebytes
from binascii import Error
from notarizer.exceptions import InvalidLabelSignature


def bytes_to_str(content: bytes, encoding="utf-8") -> str:
    return str(content, encoding=encoding)


def encode_to_base64(content: bytes) -> bytes:
    try:
        result = b64encode(content)
        return result
    except Error as err:
        raise InvalidLabelSignature(str(err))


def decode_from_base64(content: str, encoding="utf-8") -> bytes:
    try:
        result = decodebytes(str_to_bytes(content, encoding))
        return result
    except Error as err:
        raise InvalidLabelSignature(str(err))


def str_to_bytes(content: str, encoding="utf-8") -> bytes:
    return bytes(content, encoding)



import notarizer.credentials_repository as credentials_repository
import notarizer.cryptography_wrapper as crypto_wrapper
import notarizer.docker_gateway as docker_gateway
import notarizer.templates as templating

client = docker_gateway.get_client()


def sign(image: str, private_key_file_name: str, signature_label: str):
    private_key = credentials_repository.load_private_key(private_key_file_name)

    _sign_image(image, private_key, signature_label)


def _sign_image(image: str, private_key: bytes, signature_label: str):
    column_to_filter = "CreatedBy"
    history = docker_gateway.get_history(client, image, column_to_filter)
    image_history = "\n".join(history)
    signature = crypto_wrapper.create_signature(private_key, image_history)
    dockerfile_template = templating.create_signed_dockerfile(image, signature_label, signature)
    docker_gateway.build_image(client, dockerfile_template, image)



class NotarizerException(Exception):
    def __init__(self, error_code, error_message):
        super().__init__(error_message)
        self.error_code = error_code


class NoSignatureFound(NotarizerException):
    def __init__(self, error_message):
        super().__init__(10, error_message)


class InvalidLabelSignature(NotarizerException):
    def __init__(self, error_message):
        super().__init__(11, error_message)


class VerificationFailure(NotarizerException):
    def __init__(self, error_message):
        super().__init__(12, error_message)


class ImageNotFound(NotarizerException):
    def __init__(self, error_message):
        super().__init__(14, error_message)


class SigningError(NotarizerException):
    def __init__(self, reason):
        super().__init__(16, "Error Creating Signed Docker Image")
        self.reason = reason


class PublicKeyNotFound(NotarizerException):
    def __init__(self):
        super().__init__(13, "No Public Key Provided")


class PrivateKeyNotFound(NotarizerException):
    def __init__(self):
        super().__init__(15, "No Private Key Provided")



from io import BytesIO

import docker.errors
from docker import DockerClient, from_env

import notarizer.exceptions


def get_client() -> DockerClient:
    return from_env()


def get_history(client: DockerClient, image: str, column_to_filter="") -> str:
    try:
        full_history = client.api.history(image)
    except docker.errors.ImageNotFound:
        raise notarizer.exceptions.ImageNotFound("The image: {image} was not found".format(image=image))

    if column_to_filter != "":
        history = [hline[column_to_filter] for hline in full_history]
        return history

    return full_history


def build_image(client: DockerClient, dockerfile: str, image: str):
    dockerfile_in_bytes = BytesIO(bytes(dockerfile, "utf-8"))
    try:
        client.images.build(fileobj=dockerfile_in_bytes, tag=image)
    except docker.errors.BuildError as build_error:
        raise notarizer.exceptions.SigningError(str(build_error))



from setuptools import find_packages, setup
from os import getenv
from datetime import datetime

MAJOR = '0'
MINOR = '1'
PATCH = datetime.now().timetuple().tm_yday
BRANCH = getenv('CURRENT_BRANCH')
BUILD_ID = getenv('BUILD_ID')

def get_version(major, minor, patch, branch, build_id):
    version_core = f'{major}.{minor}.{patch}'
    if branch == 'main':
        return version_core

    return f'{version_core}b{build_id}'


setup_requirements = ["wheel"]
test_requirements = []

setup(
    name='notarizer',
    version=get_version(MAJOR, MINOR, PATCH, BRANCH, BUILD_ID),
    url="https://github.com/Farfetch/notarizer.git",
    license="MIT License (MIT)",
    include_package_data=True,
    packages=find_packages(exclude=["notarizer/test"]),
    setup_requires=setup_requirements,
    install_requires=open('requirements.txt').read().splitlines(),
    author="Farfetch",
    author_email="opensource@farfetch.com",
    description="Notarizer is a tool that provides a way of verifying the authenticity of docker images.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown"
)


import unittest
from unittest.mock import patch

import notarizer.converter as converter
import notarizer.credentials_repository as credentials_repository
import notarizer.cryptography_wrapper as crypto_wrapper
import notarizer.docker_gateway as docker_gateway
import notarizer.image_verifier as image_verifier
from notarizer.exceptions import (ImageNotFound, InvalidLabelSignature,
                                  NoSignatureFound, NotarizerException,
                                  PublicKeyNotFound, VerificationFailure)
from notarizer.test.mocked_data import MockedData


class TestVerifySignature(unittest.TestCase):
    def setUp(self):
        self.func = image_verifier.verify
        self.key_fake = MockedData.get_public_key_bytes()
        self.signature_name = "signature"
        self.valid_history_fake = MockedData.get_fake_valid_history_content(self.signature_name)
        self.invalid_history = MockedData.get_fake_invalid_history_content(self.signature_name)

    def test_no_signature_found_raises_no_signature_found(self):
        # Arrange
        with patch.object(
                credentials_repository,
                "load_public_key",
                return_value=self.key_fake
            ), patch.object(
                docker_gateway,
                "get_history"
            ), self.assertRaises(
                NoSignatureFound
                ) as ex:
            # Act
            self.func(["image:tag"], "key", self.signature_name)
            # Assert
            exception = ex.exception
            self.assertEqual(exception.error_code, 10)

    def test_invalid_signature_raises_invalid_image_signature(self):
        # Arrange
        with patch.object(
                credentials_repository, "load_public_key", return_value=self.key_fake
            ), patch.object(
                docker_gateway, "get_history", return_value=self.invalid_history,
            ), patch.object(
                converter, "decode_from_base64", side_effect=InvalidLabelSignature("error")
            ), self.assertRaises(
                InvalidLabelSignature
                ) as ex:
            # Act
            self.func(["image:tag"], "key", self.signature_name)
            # Assert
            exception = ex.exception
            self.assertEqual(exception.error_code, 11)

    def test_invalid_signature_raises_verification_failure(self):
        # Arrange
        with patch.object(
                credentials_repository, "load_public_key", return_value=self.key_fake
            ), patch.object(
                docker_gateway, "get_history", return_value=self.invalid_history
            ), patch.object(
                crypto_wrapper, "verify", side_effect=VerificationFailure("error")
            ), self.assertRaises(
                VerificationFailure
                ) as ex:
            # Act
            self.func(["image:tag"], "key", self.signature_name)
            # Assert
            exception = ex.exception
            self.assertEqual(exception.error_code, 12)

    def test_no_public_key_provided_raises_no_public_key_found(self):
        # Arrange
        with patch.object(
                credentials_repository, "load_public_key", side_effect=PublicKeyNotFound(),
        ), self.assertRaises(PublicKeyNotFound) as ex:
            # Act
            self.func(None, None, self.signature_name)
            # Assert
            exception = ex.exception
            self.assertEqual(exception.error_code, 13)

    def test_image_not_found_image_not_found(self):
        # Arrange
        with patch.object(
                credentials_repository, "load_public_key", return_value=self.key_fake
            ), patch.object(
                docker_gateway, "get_history", side_effect=ImageNotFound("error message"),
            ), self.assertRaises(
                ImageNotFound
                ) as ex:
            # Act
            self.func(["image:tag"], "key", self.signature_name)
            # Assert
            exception = ex.exception
            self.assertEqual(exception.error_code, 14)

    def test_valid_public_key_and_valid_history_does_not_raise_exception(self):
        # Arrange
        with patch.object(
                credentials_repository, "load_public_key", return_value=self.key_fake
            ), patch.object(
                docker_gateway, "get_history", return_value=self.valid_history_fake
                ), patch.object(crypto_wrapper, "verify"):
            # Act
            try:
                self.func(["image:tag"], "key", self.signature_name)
            except NotarizerException:
                self.fail("Should not raise an exception.")


if __name__ == "__main__":
    unittest.main()



from notarizer.exceptions import PrivateKeyNotFound, PublicKeyNotFound


def load_public_key(file_path: str) -> bytes:
    try:
        return _read_file(file_path)
    except OSError:
        raise PublicKeyNotFound()


def load_private_key(file_path: str) -> bytes:
    try:
        return _read_file(file_path)
    except OSError:
        raise PrivateKeyNotFound()


def _read_file(path: str) -> bytes:
    with open(path, "rb") as key_file:
        key_bytes = key_file.read().strip()
        return key_bytes



import sys
from typing import Sequence

import click
import notarizer.image_verifier as image_verifier
import notarizer.image_signer as image_signer
import notarizer.exceptions


@click.group()
def main():
    """
    Simple CLI for Verifying and Signing Images
    """


@main.command()
@click.option("--image", "-i", required=True, multiple=True)
@click.option("--public-key", "-p", required=True)
@click.option("--signature-label", "-s", required=False, default="signature")
def verify(image, public_key, signature_label):
    for img in image:
        image_verifier.verify(img, public_key, signature_label)
        print("{img} Verification OK".format(img=img))


@main.command()
@click.option("--image", "-i", required=True, multiple=True)
@click.option("--private-key", "-p", required=True)
@click.option("--signature-label", "-s", required=False, default="signature")
def sign(image: Sequence[str], private_key: str, signature_label: str):
    for img in image:
        image_signer.sign(img, private_key, signature_label)
        print("{img} Signature OK".format(img=img))


if __name__ == "__main__":
    try:
        main()
    except notarizer.exceptions.NotarizerException as nte:
        print(str(nte), file=sys.stderr)
        sys.exit(nte.error_code)



import unittest
from unittest.mock import patch

import notarizer.image_signer as image_signer
import notarizer.credentials_repository as credentials_repository
import notarizer.docker_gateway as docker_gateway
import notarizer.cryptography_wrapper as crypto_wrapper

from notarizer.exceptions import (
    InvalidLabelSignature,
    PrivateKeyNotFound,
    ImageNotFound,
    NotarizerException,
    SigningError
)
from notarizer.test.mocked_data import MockedData


class TestImageSigner(unittest.TestCase):
    def setUp(self):
        self.func = image_signer.sign
        self.priv_key_fake = MockedData.get_private_key_bytes()
        self.valid_history = MockedData.get_fake_valid_history_content("sign")

    def test_private_key_does_not_exist_raises_private_key_not_found(self,):
        # Arrange
        with patch.object(
                credentials_repository,
                "load_private_key",
                side_effect=PrivateKeyNotFound()
            ), self.assertRaises(
                PrivateKeyNotFound
                ) as exp:
            # Act
            self.func("test:1", "path/not_exists.pem", "dummy")

            # Assert
            self.assertEqual(exp.exception.error_code, 15)

    def test_sign__docker_image_does_not_raises_image_not_found(self):
        # Arrange
        with patch.object(credentials_repository, "load_private_key", return_value=self.priv_key_fake,), patch.object(
                docker_gateway, "get_history", side_effect=ImageNotFound("not_exists:1"),
            ), self.assertRaises(
                ImageNotFound
                ) as exp:
            # Act
            self.func("not_exists:1", "path/private.pem", "dummy")

            # Assert
            self.assertEqual(exp.exception.error_code, 14)

    def test_error_creating_signature_raises_invalid_label_signature(self):
        # Arrange
        with patch.object(credentials_repository, "load_private_key", return_value=self.priv_key_fake,), patch.object(
                docker_gateway, "get_history", side_effect=self.valid_history,
            ), patch.object(
                crypto_wrapper, "create_signature", side_effect=InvalidLabelSignature("error_msg")
            ), self.assertRaises(
                InvalidLabelSignature
                ) as exp:
            # Act
            self.func("exists:1", "path/private.pem", "dummy")

            # Assert
            self.assertEqual(exp.exception.code, 11)

    def test_sign_build_error_raises_error_creating_signed_image(self):
        # Arrange
        with patch.object(credentials_repository, "load_private_key", return_value=self.priv_key_fake,), patch.object(
                docker_gateway, "get_history", side_effect=self.valid_history,
            ), patch.object(crypto_wrapper, "create_signature", return_value="abc"), patch.object(
                docker_gateway, "build_image", side_effect=SigningError("forced error reason"),
            ), self.assertRaises(
                SigningError
                ) as exp:
            # Act
            self.func("exists:1", "path/private.pem", "dummy")

            # Assert
            self.assertEqual(exp.exception.code, 16)

    def test_sign_success_signing_image(self):
        # Arrange
        with patch.object(credentials_repository, "load_private_key", return_value=self.priv_key_fake,), patch.object(
                docker_gateway, "get_history", side_effect=self.valid_history,
            ), patch.object(
                crypto_wrapper, "create_signature", return_value="abc"
            ), patch.object(
                docker_gateway, "build_image"
                ):
            try:
                # Act
                self.func("exists:1", "path/private.pem", "dummy")
            except NotarizerException:
                self.fail("Should not raise an exception")



from notarizer.exceptions import NoSignatureFound
import notarizer.converter as converter
import notarizer.credentials_repository as credentials_repository
import notarizer.cryptography_wrapper as crypto_wrapper
import notarizer.docker_gateway as docker_gateway

client = docker_gateway.get_client()


def verify(image: str, public_key_file_name: str, signature_label: str):
    signature_line_pattern = "/bin/sh -c #(nop)  LABEL {label}=".format(label=signature_label)
    public_key = credentials_repository.load_public_key(public_key_file_name)

    column_to_filter = "CreatedBy"
    history = docker_gateway.get_history(client, image, column_to_filter)

    (signature, parent_history_start) = _extract_signature_from_history(history, signature_line_pattern, image)

    parent_image_history = "\n".join(history[i] for i in range(parent_history_start, len(history)))

    decoded_signature = converter.decode_from_base64(signature)
    crypto_wrapper.verify(public_key, decoded_signature, parent_image_history)


def _extract_signature_from_history(image_history: str, signature_line_pattern: str, image: str):
    parent_image_history_starting_line = -1
    signature = None
    for hline in image_history:
        if hline.startswith(signature_line_pattern):
            signature = hline.replace(signature_line_pattern, "")
            parent_image_history_starting_line = image_history.index(hline) + 1
            break

    if signature is None or len(signature) == 0 or parent_image_history_starting_line == -1:
        raise NoSignatureFound("No Signature Found for image: {img}. \n\
Please make sure you are using the signed parent image in your Dockerfile.".format(img=image))

    return (signature, parent_image_history_starting_line)
